package fesl

import (
	"context"
	"time"

	"gitlab.com/oiacow/goheroes/backend/core"
	"gitlab.com/oiacow/goheroes/backend/fesl/acct"
	"gitlab.com/oiacow/goheroes/backend/fesl/fsys"
	"gitlab.com/oiacow/goheroes/backend/fesl/gsum"
	"gitlab.com/oiacow/goheroes/backend/fesl/pnow"
	"gitlab.com/oiacow/goheroes/backend/fesl/rank"
	"gitlab.com/oiacow/goheroes/backend/matchmaking"
	"gitlab.com/oiacow/goheroes/backend/storage/database"
	"gitlab.com/oiacow/goheroes/backend/storage/kvstore"

	"github.com/sirupsen/logrus"
)

// Manager - handles incoming and outgoing FESL data
type Manager struct {
	db     database.Adapter
	socket *core.Socket
	server bool

	acct *acct.Account
	fsys *fsys.ConnectSystem
	gsum *gsum.GameSummary
	pnow *pnow.PlayNow
	rank *rank.Ranking
}

// New
func New(bind string, server bool, db database.Adapter, kvs *kvstore.Storage, mm *matchmaking.Pool) *Manager {
	socket, err := core.NewSocketTLS(bind)
	if err != nil {
		logrus.Fatal(err)
		return nil
	}

	fm := &Manager{
		db,
		socket,
		server,
		&acct.Account{DB: db},
		&fsys.ConnectSystem{ServerMode: server},
		&gsum.GameSummary{},
		&pnow.PlayNow{MM: mm},
		&rank.Ranking{DB: db},
	}

	return fm
}

func (fm *Manager) ListenAndServe(ctx context.Context) {
	go fm.Run(ctx)
}

func (fm *Manager) Run(ctx context.Context) {
	for {
		select {
		case event := <-fm.socket.EventChan:
			fm.handleTCP(event)
		case <-ctx.Done():
			return
		}
	}
}

func (fm *Manager) handleTCP(event core.SocketEvent) {
	ev, ok := event.Data.(core.EClientCmd)
	if !ok {
		logrus.Error("Logic error: Cant cast event to core.EClientCmd")
		return
	}

	// if !ev.Client.IsActive {
	// 	logrus.WithField("cmd", ev.Cmd).Warn("Inactive client")
	// 	return
	// }

	switch event.Name {
	case "newClient":
		fm.newClient(ev.Client) // TLS
	case "client.cmd.Hello":
		if !fm.server {
			fm.gsum.GetSessionID(ev.Client, ev.Cmd)
		}
		fm.fsys.Hello(ev.Client, ev.Cmd)
	case "client.cmd.NuLogin":
		fm.acct.NuLogin(ev)
	case "client.cmd.NuGetPersonas":
		fm.acct.NuGetPersonas(ev)
	case "client.cmd.NuGetAccount":
		fm.acct.NuGetAccount(ev)
	case "client.cmd.NuLoginPersona":
		fm.acct.NuLoginPersona(ev)
	case "client.cmd.GetStatsForOwners":
		fm.rank.GetStatsForOwners(ev)
	case "client.cmd.GetStats":
		fm.rank.GetStats(ev)
	case "client.cmd.NuLookupUserInfo":
		fm.acct.NuLookupUserInfo(ev)
	case "client.cmd.GetPingSites":
		fm.fsys.GetPingSites(ev)
	case "client.cmd.UpdateStats":
		fm.rank.UpdateStats(ev)
	case "client.cmd.TeleToken":
		fm.acct.TeleToken(ev)
	case "client.cmd.Start":
		fm.pnow.Start(ev)
		fm.pnow.Status(ev)
	case "client.cmd.MemCheck":
		// By now: Nothing interesting here, we can skip it.
		// TODO: Use MemCheck response for telemetry.
		return
	default:
		logrus.
			WithFields(logrus.Fields{
				"event":   event.Name,
				"payload": ev.Cmd.Msg,
				"query":   ev.Cmd.Query,
			}).
			Warn("fesl.UnhandledRequest")
	}
}

// TLS
func (fm *Manager) newClient(client *core.Client) {
	fm.fsys.MemCheck(client)

	client.Ticker = time.NewTicker(10 * time.Second)
	go func() {
		for {
			if !client.IsActive {
				return
			}
			select {
			case <-client.Ticker.C:
				if !client.IsActive {
					return
				}
				fm.fsys.MemCheck(client)
			}
		}
	}()

	logrus.Debug("New client has connected")
}
