package main

// Generate setters and getters for stats
// $ go generate ./cmd/goheroes
//go:generate go run ../stats-codegen/main.go -scan ../../backend/model --getters ../../backend/ranking/getters.go --setters ../../backend/ranking/setters.go --adders ../../backend/ranking/adders.go

import (
	"context"
	"flag"

	"github.com/google/gops/agent"
	"github.com/sirupsen/logrus"
	"github.com/subosito/gotenv"

	"gitlab.com/oiacow/goheroes/backend/config"
	"gitlab.com/oiacow/goheroes/backend/core"
	"gitlab.com/oiacow/goheroes/backend/fesl"
	"gitlab.com/oiacow/goheroes/backend/matchmaking"
	"gitlab.com/oiacow/goheroes/backend/storage/database"
	"gitlab.com/oiacow/goheroes/backend/storage/kvstore"
	"gitlab.com/oiacow/goheroes/backend/theater"
)

func main() {
	initConfig()
	initLogger()

	ctx, cancel := context.WithCancel(context.Background())
	defer func() {
		cancel()
	}()
	startServer(ctx)

	// Use "github.com/google/gops/agent"
	if err := agent.Listen(agent.Options{}); err != nil {
		logrus.Fatal(err)
	}

	logrus.Println("Serving...")
	<-ctx.Done()
}

func initConfig() {
	var (
		configFile string
	)
	flag.StringVar(&configFile, "config", ".env", "Path to confi file")
	flag.Parse()

	gotenv.Load(configFile)
	config.Initialize()
}

func initLogger() {
	logrus.SetLevel(config.LogLevel())

	logrus.SetFormatter(&logrus.JSONFormatter{
		DisableTimestamp: true,
	})
}

func startServer(ctx context.Context) {
	db, err := database.New()
	if err != nil {
		logrus.Fatal(err)
	}

	core.InitClientData()
	kvs := kvstore.NewInMemory()
	mm := matchmaking.NewPool()

	fesl.New(config.FeslCliAddr(), false, db, kvs, mm).ListenAndServe(ctx)
	fesl.New(config.FeslServAddr(), true, db, kvs, mm).ListenAndServe(ctx)

	theater.New(config.TmCliAddr(), db, kvs, mm).ListenAndServe(ctx)
	theater.New(config.TmServAddr(), db, kvs, mm).ListenAndServe(ctx)
}
